# -*- coding: utf-8 -*-

from aiohttp import ClientSession, TCPConnector, ClientTimeout
import asyncio
from cachetools import TTLCache
from collections import defaultdict
from functools import update_wrapper
from requests import Session

from .api_toolkit import (
    default_headers,
    isiter_noliterals,
    ismapping,
    multiparams,
)
from .base_classes import AsyncInitObject, AsyncWith, SyncWith
from .cache_utils import somecachedmethod, iscorofunc
from .exceptions import WITH_CODE, UnexpectedResponseCode
from .typedefs import URLS, L, R

from typing import (
    Any,
    Coroutine,
    Dict,
    Generator,
    Generic,
    Iterable,
    List,
    Mapping,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import orjson

__all__ = (
    "AsyncSession",
    "SyncSession",
)


# XXX: in both functions I need to find a suitable cache_limit
# 1024 is a relatively random choice and
# has nothing to do with the desired behavior

def _raise_for_status(self, url: str, code: int,
                      data: Union[Mapping[str, Any], str]) -> None:

    if isinstance(data, str):
        reason = "without a reason"
        message = data if len(data) > 0 else "no message"
    else:
        reason = data.get("reason", "without a reason")
        message = data.get("message", "no message")

    excp = next(filter(lambda x: x.code == code, WITH_CODE), None)
    if excp is not None:
        raise excp(url, reason, message)
    else:
        raise UnexpectedResponseCode(url, code, reason, message)


def mix_all_gets(multi):
    def decorator(func):
        coro = iscorofunc(func)

        _cached_get = somecachedmethod(func)

        if coro:
            async def no_cache(self, *args, **kwrags):
                self._last_urls.append(args[0])
                res = await func(self, *args, **kwrags)
                self._last_reqs.append((args, kwrags, res))

            async def _cache(self, *args, **kwrags):
                self._last_urls.append(args[0])
                res = await _cached_get(self, *args, **kwrags)
                self._last_reqs.append((args, kwrags, res))
        else:
            def no_cache(self, *args, **kwrags):
                self._last_urls.append(args[0])
                res = func(self, *args, **kwrags)
                self._last_reqs.append((args, kwrags, res))

            def _cache(self, *args, **kwrags):
                self._last_urls.append(args[0])
                res = _cached_get(self, *args, **kwrags)
                self._last_reqs.append((args, kwrags, res))

        _multi_get = multiparams(no_cache)
        _cached_multi_get = multiparams(_cache)

        if coro and multi:
            async def wrapper(self, *args: Any, **kwrags: Any):
                if self.can_use_cache:
                    await _cached_multi_get(self, *args, **kwrags)
                else:
                    await _multi_get(self, *args, **kwrags)
        elif not coro and multi:
            def wrapper(self, *args: Any, **kwrags: Any):
                if self.can_use_cache:
                    _cached_multi_get(self, *args, **kwrags)
                else:
                    _multi_get(self, *args, **kwrags)
        elif coro and not multi:
            async def wrapper(self, *args: Any, **kwrags: Any):
                if self.can_use_cache:
                    await _cache(self, *args, **kwrags)
                else:
                    await no_cache(self, *args, **kwrags)
        else:
            def wrapper(self, *args: Any, **kwrags: Any):
                if self.can_use_cache:
                    _cache(self, *args, **kwrags)
                else:
                    no_cache(self, *args, **kwrags)
        return wrapper
    return decorator


def retry_to_get_data(func):
    rte = RuntimeError(
        "self._attempts argument was changed"
        " causing it to work incorrectly")

    if iscorofunc(func):
        async def wrapper(self, *args, **kwrags):
            good_resps = defaultdict(list)
            d_kwargs = defaultdict(list)
            d_args = defaultdict(list)
            for i in self._attempts:
                if len(self._last_reqs) == 0:
                    in_args, in_kwrags = args, kwrags
                else:
                    in_args, in_kwrags = d_args.values(), d_kwargs

                self._last_reqs.clear()
                self._last_urls.clear()

                await func(self, *in_args, **in_kwrags)

                d_args.clear()
                d_kwargs.clear()

                for a, kw, (code, data) in self._last_reqs:
                    url = a[0]
                    if code == 200:
                        good_resps[url].append(data)
                    elif i == 0:
                        self.raise_for_status(url, code, data)
                    else:
                        for key, val in kw.items():
                            d_kwargs[key].append(val)

                        for i, val in enumerate(a):
                            d_args[i].append(val)

                if len(d_args) == len(d_kwargs) == 0:
                    ret = [good_resps[url].pop(0) for url in self._last_urls]
                    self._last_reqs.clear()
                    self._last_urls.clear()
                    return ret

            raise rte
    else:
        def wrapper(self, *args, **kwrags):
            good_resps = defaultdict(list)
            d_kwargs = defaultdict(list)
            d_args = defaultdict(list)
            for i in self._attempts:
                if len(self._last_reqs) == 0:
                    in_args, in_kwrags = args, kwrags
                else:
                    in_args, in_kwrags = d_args.values(), d_kwargs

                self._last_reqs.clear()
                self._last_urls.clear()

                func(self, *in_args, **in_kwrags)

                d_args.clear()
                d_kwargs.clear()

                for a, kw, (code, data) in self._last_reqs:
                    url = a[0]
                    if code == 200:
                        good_resps[url].append(data)
                    elif i == 0:
                        self.raise_for_status(url, code, data)
                    else:
                        for key, val in kw:
                            d_kwargs[key].append(val)

                        for i, val in enumerate(a):
                            d_args[i].append(val)

                if len(d_args) == len(d_kwargs) == 0:
                    ret = [good_resps[url].pop(0) for url in self._last_urls]
                    self._last_reqs.clear()
                    self._last_urls.clear()
                    return ret

            raise rte

    return update_wrapper(wrapper, func)


def headers_handler(self, headers):
    if not self.can_use_cache:
        return headers

    if isiter_noliterals(headers) and not ismapping(headers):
        return [tuple(zip(headers.items())) for header in headers]
    else:
        return tuple(zip(headers.items()))


class AsyncSession(AsyncInitObject, AsyncWith):
    async def __init__(self, trust_env: bool = True,
                       cache_ttl: Union[int, float] = 60,
                       cache_limit: int = 1024,
                       use_cache: bool = True,
                       timeout: Union[int, float] = 30,
                       repeat_failed: int = 3) -> None:
        headers = default_headers()
        loop = asyncio.get_event_loop()
        self.session = ClientSession(
            loop=loop,
            connector=TCPConnector(use_dns_cache=False, loop=loop),
            trust_env=trust_env,
            headers=headers,
            timeout=ClientTimeout(total=timeout),
        )

        if use_cache:
            self.cache = TTLCache(maxsize=cache_limit, ttl=cache_ttl)
        else:
            self.cache = None
        self._use_cache = use_cache

        if repeat_failed > 1:
            self._attempts = range(repeat_failed - 1, -1, -1)
        else:
            self._attempts = (0)

        self._last_reqs = []
        self._last_urls = []

    async def close(self) -> None:
        """Close underlying connector.
        Release all acquired resources.
        """
        if not self.closed:
            # XXX: https://github.com/aio-libs/aiohttp/issues/1925
            # https://docs.aiohttp.org/en/stable/client_advanced.html#graceful-shutdown
            await self.session.close()
            # Zero-sleep to allow underlying connections to close
            await asyncio.sleep(0.300)

    @property
    def closed(self) -> bool:
        """Is client session closed.
        A readonly property.
        """
        return self.session.closed

    raise_for_status = _raise_for_status

    @property
    def can_use_cache(self) -> bool:
        return self._use_cache and isinstance(self.cache, TTLCache)

    async def _simple_get(
            self, url: str, from_json: bool = True,
            headers: Iterable[Iterable[str]] = {}) -> Tuple[int, str]:
        async with self.session.get(url, headers=dict(headers)) as response:
            code = response.status
            data = await response.text()
            if from_json:
                data = orjson.loads(data)

        return code, data

    _get = retry_to_get_data(mix_all_gets(False)(_simple_get))
    _gets = retry_to_get_data(mix_all_gets(True)(_simple_get))

    async def get(self, url: str, from_json: bool = True,
                  headers: Dict[str, str] = {}) -> R:
        return (await self._get(
            url, from_json=from_json,
            headers=headers_handler(self, headers)))[0]

    async def gets(
            self, urls: URLS, from_jsons: bool = True,
            headers: Union[List[Dict[str, str]], Dict[str, str]] = {}) -> L:
        return await self._gets(
            urls, from_json=from_jsons,
            headers=headers_handler(self, headers))


class SyncSession(SyncWith):
    def __init__(self, trust_env: bool = True,
                 cache_ttl: Union[int, float] = 60,
                 cache_limit: int = 1024, use_cache: bool = True,
                 timeout: Union[int, float] = 30,
                 repeat_failed: int = 3) -> None:
        self._closed = False

        headers = default_headers()
        self.session = Session()
        self.session.trust_env = trust_env
        self.session.headers.update(headers)

        if use_cache:
            self.cache = TTLCache(maxsize=cache_limit, ttl=cache_ttl)
        else:
            self.cache = None
        self._use_cache = use_cache

        self.timeout = timeout

        if repeat_failed > 1:
            self._attempts = range(repeat_failed - 1, -1, -1)
        else:
            self._attempts = (0)

        self._last_reqs = []
        self._last_urls = []

    def close(self) -> None:
        """Closes all adapters and as such the session"""
        if not self.closed:
            self.session.close()
            self._closed = True

    @property
    def closed(self) -> bool:
        """Is client session closed.
        A readonly property.
        """
        return self._closed

    raise_for_status = _raise_for_status

    @property
    def can_use_cache(self) -> bool:
        return self._use_cache and isinstance(self.cache, TTLCache)

    def _simple_get(
            self, url: str, from_json: bool = True,
            headers: Iterable[Iterable[str]] = {}) -> Tuple[int, str]:
        with self.session.get(
                url, timeout=self.timeout, headers=dict(headers)) as response:
            code = response.status_code
            data = response.text
            if from_json:
                data = orjson.loads(data)

        return code, data

    _get = retry_to_get_data(mix_all_gets(False)(_simple_get))
    _gets = retry_to_get_data(mix_all_gets(True)(_simple_get))

    def get(self, url: str, from_json: bool = True,
            headers: Dict[str, str] = {}) -> R:
        return self._get(url, from_json=from_json,
                         headers=headers_handler(self, headers))[0]

    def gets(self, urls: URLS, from_jsons: bool = True,
             headers: Union[List[Dict[str, str]], Dict[str, str]] = {}) -> L:
        return self._gets(urls, from_json=from_jsons,
                          headers=headers_handler(self, headers))



# -*- coding: utf-8 -*-

import asyncio
from cachetools import keys, Cache
from functools import wraps, partial, update_wrapper

__all__ = (
    "NaN",
    "async_cachedmethod",
    "cachedmethod",
    "iscorofunc",
    "get_decorator",
    "somecachedmethod",
    "classcache"
)


class NaN:
    "NaN"
    pass


def async_cachedmethod(key=keys.hashkey, lock=None):
    """Decorator to wrap a class or instance method with a memoizing
    callable that saves results in a cache.
    """
    def decorator(method):
        if lock is None:
            async def wrapper(self, *args, **kwargs):
                k = key(*args, **kwargs)
                cache = self.cache
                get_k = cache.get(k, NaN)
                if get_k != NaN:
                    return get_k
                v = await method(self, *args, **kwargs)
                try:
                    cache[k] = v
                except ValueError:
                    pass  # value too large
                return v
        else:
            async def wrapper(self, *args, **kwargs):
                k = key(*args, **kwargs)
                cache = self.cache
                with lock(self):
                    get_k = cache.get(k, NaN)
                if get_k != NaN:
                    return get_k
                v = await method(self, *args, **kwargs)
                try:
                    with lock(self):
                        cache[k] = v
                except ValueError:
                    pass  # value too large
                return v
        return update_wrapper(wrapper, method)
    return decorator


def cachedmethod(key=keys.hashkey, lock=None):
    """Decorator to wrap a class or instance method with a memoizing
    callable that saves results in a cache.
    """
    def decorator(method):
        if lock is None:
            def wrapper(self, *args, **kwargs):
                k = key(*args, **kwargs)
                cache = self.cache
                get_k = cache.get(k, NaN)
                if get_k != NaN:
                    return get_k
                v = method(self, *args, **kwargs)
                try:
                    cache[k] = v
                except ValueError:
                    pass  # value too large
                return v
        else:
            def wrapper(self, *args, **kwargs):
                k = key(*args, **kwargs)
                cache = self.cache
                with lock(self):
                    get_k = cache.get(k, NaN)
                if get_k != NaN:
                    return get_k
                v = method(self, *args, **kwargs)
                try:
                    with lock(self):
                        cache[k] = v
                except ValueError:
                    pass  # value too large
                return v
        return update_wrapper(wrapper, method)
    return decorator


def iscorofunc(func):
    return asyncio.iscoroutinefunction(func)


def iscoro(func):
    return asyncio.iscoroutine(func)


def get_decorator(func):
    if iscorofunc(func):
        return async_cachedmethod
    else:
        return cachedmethod


def somecachedmethod(func):
    # key = partial(keys.hashkey, func.__name__)
    return get_decorator(func)()(func)  # key=key



# -*- coding: utf-8 -*-

from .api_toolkit import make_headers
from .optional_format import optional_format

from typing import Any, Dict, Optional, Union
import urllib.parse as parse

__all__ = (
    "API",
    "api_defs"
)


class API:
    def __init__(self, base: str, paths: Dict[str, str] = {},
                 params: Dict[str, Dict[str, str]] = {},
                 token: Optional[str] = None) -> None:
        if not (base.startswith("http://") or base.startswith("https://")):
            base = "https://" + base
        elif base.startswith("http://"):
            base = "https://" + base[len("http://"):]

        if not base.endswith("/"):
            base += "/"

        if len(set(params) - set(paths)) != 0:
            raise ValueError(
                "'params.keys()' must be in the 'paths.keys()'")

        for name, path in paths.items():
            paths[name] = parse.urljoin(base, path)

            if params.get(name) is None:
                params[name] = {}

        self.base = base
        self.paths = paths
        self.params = params
        self.set_token(token)

    def __getattr__(self, name):
        get = self.paths.get(name)
        if get is None:
            raise AttributeError(f"'API' object has no attribute '{name}'")

        return get

    def set_token(self, token: str):
        if token is None:
            self.headers = {}
        else:
            self.headers = make_headers(token)

    def get(self, name: str, **params: Any) -> str:
        url = getattr(self, name)
        if len(params) == 0:
            return url

        for_format = {}
        used_parameters = []

        for param, val in params.items():
            get = self.params[name].get(param)
            if get is None:
                for_format.update({param: val})
            elif get != val:
                used_parameters.append(f"{param}={val}")

        if len(used_parameters) != 0:
            url += "?" + "&".join(used_parameters)

        return optional_format(url, **params)


official = {
    "players": "players/{tag}",
    "battlelog": "players/{tag}/battlelog",
    "clubs": "clubs/{tag}",
    "members": "clubs/{tag}/members",
    "rankings": "rankings/{code}/{kind}/{id_}",
    "brawlers": "brawlers/{id_}",
}

# before and after - this is so impractical that I suppose nobody will use this
# that's why I decided not to include it here
offic_params = {
    "members": {"limit": "100"},
    "rankings": {"limit": "200"},
    "brawlers": {"limit": ""},
}

starlist = {
    "events": "events",
    "brawlers": "brawlers",
    "icons": "icons",
    "maps": "maps/{id_}",
    "gamemodes": "gamemodes",
    "clublog": "clublog/{tag}",
}

api_defs = {
    "official": API("api.brawlstars.com/v1", official, offic_params),
    "starlist": API("api.starlist.pro", starlist),
}



# -*- coding: utf-8 -*-

from . import __version__, __name__
from .cache_utils import somecachedmethod, iscorofunc
from asyncio import ensure_future as ensure, gather
from collections.abc import Iterable, ByteString, Mapping
from functools import update_wrapper
import sys
from typing import Dict, Union

__all__ = (
    "default_headers",
    "make_headers",
    "isiter_noliterals",
    "isunit",
    "isempty",
    "ismapping",
    "same",
    "unique",
    "check_kwargs",
    "check_args",
    "check_params",
    "_rearrange_params",
    "rearrange_params",
    "multiparams",
    "multiparams_classcache"
)


def default_headers() -> Dict[str, str]:
    return {
        "dnt": "1",
        "user-agent": f"{__name__}/{__version__} (Python {sys.version[:5]})",
        "accept-encoding": ", ".join(("gzip", "deflate")),
        "cache-control": "no-cache",
        "pragma": "no-cache",
    }


def make_headers(token: str) -> Dict[str, str]:
    return {"authorization": f"Bearer {token}"}


def isiter_noliterals(obj):
    return isinstance(obj, Iterable) and not isinstance(obj, (str, ByteString))


def isunit(obj):
    return len(obj) == 1
    # isinstance(obj, tuple) and .


def isempty(obj):
    return len(obj) == 0


def ismapping(obj):
    return isinstance(obj, Mapping)


def isrequiredtype(obj):
    return (
        isiter_noliterals(obj) and not ismapping(obj)
        and not isempty(obj))


def same(elements):
    return len(elements) == elements.count(elements[0])


def unique(x):
    seen = list()
    return not any(i in seen or seen.append(i) for i in x)


def check_kwargs(kwargs):
    lengths = []
    for key, param in kwargs.items():
        p_type = type(param)
        if isrequiredtype(param):
            if isunit(param):
                kwargs[key] = ("u", param[0], p_type)
            else:
                lengths.append(len(param))
                kwargs[key] = ("m", iter(param), p_type)
        else:
            kwargs[key] = ("u", param, p_type)

    return kwargs, lengths


def check_args(args):
    args = list(args)
    lengths = []
    for i, param in enumerate(args):
        p_type = type(param)
        if isrequiredtype(param):
            if isunit(param):
                args[i] = ("u", param[0], p_type)
            else:
                lengths.append(len(param))
                args[i] = ("m", iter(param), p_type)
        else:
            args[i] = ("u", param, p_type)

    return args, lengths


def check_params(args, kwargs):
    all_args, args_lengths = check_args(args[:])
    all_kwargs, kwargs_lengths = check_kwargs(kwargs.copy())

    lengths = args_lengths + kwargs_lengths

    if len(lengths) < 1:
        total_length = 1
    else:
        if not same(lengths):
            raise ValueError(
                "All allowed iterable parameters must be of the same length.")

        total_length = lengths[0]

    return all_args, all_kwargs, total_length


def _rearrange_params(args, kwargs):
    all_args, all_kwargs, length = check_params(args[:], kwargs.copy())

    for _ in range(length):
        new_args = []
        for (kind, val, p_type) in all_args:
            if kind == "m":
                val = next(val)
                #val = p_type([val])
            new_args.append(val)

        new_kwargs = {}
        for key, (kind, val, p_type) in all_kwargs.items():
            if kind == "m":
                val = next(val)
                #val = p_type([val])
            new_kwargs[key] = val

        yield new_args, new_kwargs


def rearrange_params(*args, **kwargs):
    return _rearrange_params(args, kwargs)


def multiparams(func):
    if iscorofunc(func):
        async def wrapper(*args, **kwargs):
            params = _rearrange_params(args, kwargs)
            tasks = [ensure(func(*a, **kw)) for a, kw in params]
            return await gather(*tasks)
    else:
        def wrapper(*args, **kwargs):
            params = _rearrange_params(args, kwargs)
            return [func(*a, **kw) for a, kw in params]
    return update_wrapper(wrapper, func)



# -*- coding: utf-8 -*-

from typing import Any

__all__ = (
    "ClientException",

    "ClientResponseError",
    "UnexpectedResponseCode",
    "BadRequest", "Forbidden", "NotFound",
    "TooManyRequests", "InternalServerError",
    "ServiceUnavailable", "WITH_CODE",
)


class ClientException(Exception):
    """Base class for all client exceptions."""


class ClientResponseError(ClientException):
    """Connection error during reading response."""

    def __init__(self, url: str, reason: str = "without a reason",
                 message: str = "no message"):
        self.url = url
        self.reason = reason
        self.message = message
        self.cause = "because " if reason != "without a reason" else ""

    def __repr__(self):
        return (
            "{0.__class__.__name__}({0.url!r}, "
            "{0.reason!r}, {0.message!r})").format(self)

    def __str__(self):
        return (
            "{0.url} -> {0.code}, {0.cause}"
            "{0.reason}, {0.message}").format(self)

    def __eq__(self, other):
        return (
            (dir(self) == dir(other))
            and (vars(self) == vars(other)))


class UnexpectedResponseCode(ClientResponseError):
    """Occurs if the response code was not described
    in the official api documentation.
    """

    def __init__(self, url: str, code: int, *args: Any, **kwargs: Any):
        self.code = code
        super().__init__(url, *args, **kwargs)

    def __repr__(self):
        return (
            "{0.__class__.__name__}({0.url!r}, {0.code!r}, "
            "{0.reason!r}, {0.message!r})").format(self)


class BadRequest(ClientResponseError):
    """Client provided incorrect parameters for the request."""

    code = 400


class Forbidden(ClientResponseError):
    """Access denied, either because of missing/incorrect credentials or
    used API token does not grant access to the requested resource.
    """

    code = 403


class NotFound(ClientResponseError):
    """Resource was not found."""

    code = 404


class TooManyRequests(ClientResponseError):
    """Request was throttled, because amount of requests
    was above the threshold defined for the used API token.
    """

    code = 429


class InternalServerError(ClientResponseError):
    """Unknown error happened when handling the request."""

    code = 500


class ServiceUnavailable(ClientResponseError):
    """Service is temprorarily unavailable because of maintenance."""

    code = 503


WITH_CODE = [
    BadRequest, Forbidden, NotFound,
    TooManyRequests, InternalServerError, ServiceUnavailable
]



# -*- coding: utf-8 -*-

import string

__all__ = (
    "optional_format",
    "OptionalFormatter"
)


class OptionalFormatter(string.Formatter):
    def _vformat(self, format_string, args, kwargs, used_args, recursion_depth,
                 auto_arg_index=0):
        if recursion_depth < 0:
            raise ValueError('Max string recursion exceeded')
        result = []
        for literal_text, field_name, format_spec, conversion in \
                self.parse(format_string):

            # output the literal text
            if literal_text:
                result.append(literal_text)

            # if there's a field, output it
            if field_name is not None:
                # this is some markup, find the object and do
                #  the formatting

                self.field_name_existed = True
                # handle arg indexing when empty field_names are given.
                if field_name == '':
                    if auto_arg_index is False:
                        raise ValueError('cannot switch from manual field '
                                         'specification to automatic field '
                                         'numbering')
                    field_name = str(auto_arg_index)
                    self.field_name_existed = False
                    auto_arg_index += 1
                elif field_name.isdigit():
                    if auto_arg_index:
                        raise ValueError('cannot switch from manual field '
                                         'specification to automatic field '
                                         'numbering')
                    # disable auto arg incrementing, if it gets
                    # used later on, then an exception will be raised
                    auto_arg_index = False

                # given the field_name, find the object it references
                #  and the argument it came from
                obj, arg_used = self.get_field(field_name, args, kwargs)
                used_args.add(arg_used)

                # do any conversion on the resulting object
                obj = self.convert_field(obj, conversion)

                # expand the format spec, if needed
                format_spec, auto_arg_index = self._vformat(
                    format_spec, args, kwargs,
                    used_args, recursion_depth-1,
                    auto_arg_index=auto_arg_index)

                # format the object and append to the result
                result.append(self.format_field(obj, format_spec))

        return ''.join(result), auto_arg_index

    def get_value(self, key, args, kwargs):
        if isinstance(key, int):
            if len(args) > key:
                return args[key]
        else:
            get = kwargs.get(key)
            if get is not None:
                return get

        if self.field_name_existed:
            return f"{{{key}}}"
        else:
            return "{}"


optional_format = OptionalFormatter().format



# -*- coding: utf-8 -*-

from typing import TypeVar

__all__ = (
    "R",
    "L",
    "URLS",
    "PARAMS",
)

R = TypeVar("R", bound="Union[Dict[str, Any], List[Dict[str, Any]]]")

L = TypeVar("L", bound="List[Union[Dict[str, Any], List[Dict[str, Any]]]]")

URLS = TypeVar("URLS", bound="Union[List[str], str]")

PARAMS = TypeVar(
    "PARAMS", bound="Sequence[Sequence[Sequence[Any], Mapping[Any, Any]]]")
